;;; -*- Mode: Lisp; Package: Lisp; Log: code.log -*-
;;;
;;; **********************************************************************
;;; This code was written as part of the CMU Common Lisp project at
;;; Carnegie Mellon University, and has been placed in the public domain.
;;;
(ext:file-comment
  "$Header: src/contrib/contrib.lisp $")

;;; Empty for backward compatibility for users who require'd
;;; cmu-contribs to get the CMU contribs.  The contribs all use asdf
;;; now, so we don't need this file to do anything.



